with Messages;
use Messages;

package body Taylors.Ops is

   Half3:   constant Rep := One+Half;
   Quarter: constant Rep := Half*Half;

  procedure Neg(P: in out Taylor) is
  begin
    for K in Power loop
      P(K) := -P(K);
    end loop;
  end Neg;

  function "-"(P: Taylor) return Taylor is
    Q: Taylor;
  begin
    for K in Power loop
      Q(K) := -P(K);
    end loop;
    Q(Size) := P(Size);
    return Q;
  end "-";

  procedure Mult(S: in Scalar; P: in out Taylor) is
  begin
    for K in Power loop
      P(K) := S*P(K);
    end loop;
    if Trunc then return; end if;
    P(Size) := abs(S)*P(Size);
  end Mult;

  procedure Prod(S: in Scalar; P1: in Taylor; P2: out Taylor) is
  begin
    for K in Power loop
      P2(K) := S*P1(K);
    end loop;
    if Trunc then return; end if;
    P2(Size) := abs(S)*P1(Size);
  end Prod;

  pragma Inline (Prod);

  function "*"(S: Scalar; P: Taylor) return Taylor is
    Q: Taylor;
  begin
    Prod(S,P,Q);
    return Q;
  end "*";

  procedure Add(P1: in Taylor; P2: in out Taylor) is
  begin
    for K in Power loop
      P2(K) := P2(K)+P1(K);
    end loop;
    if Trunc then return; end if;
    P2(Size) := P2(Size)+P1(Size);
  end Add;

  procedure Sum(P1,P2: in Taylor; P3: out Taylor) is
  begin
    for K in Power loop
      P3(K) := P1(K)+P2(K);
    end loop;
    if Trunc then return; end if;
    P3(Size) := P1(Size)+P2(Size);
  end Sum;

  pragma Inline (Sum);

  function "+"(P1,P2: Taylor) return Taylor is
    Q: Taylor;
  begin
    Sum(P1,P2,Q);
    return Q;
  end "+";

  procedure MultAdd(S: in Scalar; P1: in Taylor; P2: in out Taylor) is
  begin
    for K in Power loop
      P2(K) := P2(K)+S*P1(K);
    end loop;
    if Trunc then return; end if;
    P2(Size) := P2(Size)+abs(S)*P1(Size);
  end MultAdd;

  procedure Sub(P1: in Taylor; P2: in out Taylor) is
  begin
    for K in Power loop
      P2(K) := P2(K)-P1(K);
    end loop;
    if Trunc then return; end if;
    P2(Size) := P2(Size)+P1(Size);
  end Sub;

  procedure Diff(P1,P2: in Taylor; P3: out Taylor) is
  begin
    for K in Power loop
      P3(K) := P1(K)-P2(K);
    end loop;
    if Trunc then return; end if;
    P3(Size) := P1(Size)+P2(Size);
  end Diff;

  pragma Inline (Diff);

  function "-"(P1,P2: Taylor) return Taylor is
    Q: Taylor;
  begin
    Diff(P1,P2,Q);
    return Q;
  end "-";

  procedure Mult(R: in Radius; P1: in Taylor; P2: in out Taylor) is
    S: Scalar;
  begin
    if not Trunc then
      S := ScalZero;
      for K in reverse Size .. 2*Size loop
        S := R*S;
        for J in K-Size .. Size loop
          S := S+abs(P1(K-J)*P2(J));
        end loop;
      end loop;
      P2(Size) := S;
    end if;

    for K in reverse Power loop
      S := P2(K);
      for J in 0 .. K loop
        S := S+P1(K-J)*P2(J);
      end loop;
      P2(K) := S;
    end loop;
  end Mult;

  procedure Prod(R: in Radius; P1,P2: in Taylor; P3: out Taylor) is
    S: Scalar;
  begin
    for K in reverse Power loop
      S := ScalZero;
      for J in 0 .. K loop
        S := S+P1(K-J)*P2(J);
      end loop;
      P3(K) := S;
    end loop;
    if Trunc then return; end if;

    S := ScalZero;
    for K in reverse Size .. 2*Size loop
      S := R*S;
      for J in K-Size .. Size loop
        S := S+abs(P1(K-J)*P2(J));
      end loop;
    end loop;
    P3(Size) := S;
  end Prod;

  procedure MultAdd(R: in Radius; P1,P2: in Taylor; P3: in out Taylor) is
    S: Scalar;
  begin
    for K in reverse Power loop
      S := P3(K);
      for J in 0 .. K loop
        S := S+P1(K-J)*P2(J);
      end loop;
      P3(K) := S;
    end loop;
    if Trunc then return; end if;

    S := ScalZero;
    for K in reverse Size .. 2*Size loop
      S := R*S;
      for J in K-Size .. Size loop
        S := S+abs(P1(K-J)*P2(J));
      end loop;
    end loop;
    P3(Size) := S+P3(Size);
  end MultAdd;

  function "abs"(P: Taylor) return Taylor is
     Q: Taylor;
  begin
     for J in Power loop
        Q(J) := abs(P(J));
     end loop;
     Q(Size) := P(Size);
     return Q;
  end "abs";

  function Val(R: Radius; S: Scalar; P: Taylor) return Scalar is
    V: Scalar;
  begin
    if Trunc then
      V := ScalZero;
    else
      V := Ball(P(Size));
      if SupAbs(S)>R then
        Message("Taylors.Ops.Val",Domain_Error);
      end if;
    end if;
    for K in reverse Power loop
      V := S*V+P(K);
    end loop;
    return V;
  end Val;

  function Norm(R: Radius; P: Taylor) return Scalar is
    V: Scalar;
  begin
    if Trunc then V := ScalZero; else V := Norm(P(Size)); end if;
    for K in reverse Power loop
      V := R*V+abs(P(K));
    end loop;
    return V;
  end Norm;

  function ExpTrunc(N: Integer; L: Scalar) return Taylor is
     P: Taylor;
     K: Integer := 1;
  begin
     P(0) := ScalOne;
     for J in 1 .. PDeg loop
        if J <= N then
           P(J) := P(J-1)*L/Rep(J);
        else
           P(J) := ScalZero;
        end if;
    end loop;
    P(Size) := ScalZero;
    return P;
  end ExpTrunc;

  function Exp(R: Radius; S: Taylor) return Taylor is
     TOne: Taylor  := CMonom(0,ScalOne);
     Iter: constant Integer := 15;
     K:   Integer := 0;
     M:   Rep     := Sup(Norm(R,S));
     X,Y,XY: Taylor;
  begin
    if IsZero(S) then return TOne; end if;      -- S=0
    while M>Quarter loop
       K := K+1;
       M := M*Half;
    end loop;
    X := Scal(Half**K)*S;
    Y := CMonom(0,Scal(Half3,-Half3));
    for I in reverse 1 .. Iter loop
       Prod(R,X,Y,XY);
       Y := TOne+(Scal(One)/Scal(I))*XY;
    end loop;
    for I in 1 .. K loop
       Prod(R,Y,Y,X);
       Y := X;
    end loop;
    return Y;
  end Exp;

  function D(R1,R2: Radius; P: Taylor) return Taylor is
     Q: Taylor;
     CRR: Scalar := Scal(One)/(Exp(Scal(One))*Scal(R2)*Log(Scal(R1)/Scal(R2)));
  begin
     for K in Power loop
        if K < PDeg then
           Q(K) := Scal(K+1)*P(K+1);
        else
           Q(K) := Sup(P(Size))*Scal(Rep(PDeg),-Rep(PDeg));
        end if;
     end loop;
     if Trunc then return Q; end if;
     Q(Size) := (Scal(Pdeg)/Scal(R2)+CRR)*abs(P(Size));
     return Q;
  end D;

  function Dm1(R: Radius; P: Taylor) return Taylor is
     Q: Taylor;
  begin
     Q(0) := Scal(Zero);
     for K in 1 .. PDeg loop
        Q(K) := P(K-1)/Scal(K);
     end loop;
     if Trunc then return Q; end if;
     Q(Size) := abs(P(Pdeg))/Scal(Size)+P(Size)*Scal(R)/Scal(PDeg+2);
     return Q;
  end Dm1;

  function Ball(R: Radius) return Taylor is
     P: Taylor;
     C: Scalar := Scal(One,-One)/Scal(R);
  begin
     if Trunc then return CMonom(0,ScalZero); end if;
     P(0) := ScalOne;
     for K in 1 .. PDeg loop
        P(K) := P(K-1)*C;
     end loop;
     P(Size) := abs(P(Size-1)*C);
     return P;
  end Ball;

end Taylors.Ops;
